import numpy as np
import copy
import scipy.special as sps


def single_run_adaptive_steps(pb, x, Sy, prox_y, prox_y_cpy, rf,
                              rhx, rQ, primal_step_size, dual_step_size, sm_gap_init, sm_gap, change_in_x, change_in_y,
                              nb_iterations, memory, history=None, alpha=2):

#rho_ref, sm_gap_init, sm_gap, sm_gap2, nb_iterations,
#                        tau, sigma, up_down, alpha=2, history=None):
    # Idea based on the assumption that |x - x*|^2/tau = |y - y*|^2/sigma
    """
    if sm_gap_x < 1.1 * sm_gap_x2 or sm_gap_y < 1.1 * sm_gap_y2 \
       or sm_gap_x2 < 1.1 * sm_gap or sm_gap_y2 < 1.1 * sm_gap:
        # sign of a transient regime
        return tau, sigma
    else:
        sm_gap = 0  # slow down the process
        ratio_taumufhat_sigmamughat = (sm_gap_x-sm_gap)/(sm_gap_y-sm_gap)
        ratio_taumufhat_sigmamughat2 = (sm_gap_x2-sm_gap)/(sm_gap_y2-sm_gap)

        if np.abs(np.log(ratio_taumufhat_sigmamughat)) > np.abs(np.log(ratio_taumufhat_sigmamughat2)):
            ratio_taumufhat_sigmamughat = ratio_taumufhat_sigmamughat2
        
        tau_ = tau * np.sqrt(1/ratio_taumufhat_sigmamughat)**alpha
        sigma_ = sigma * np.sqrt(ratio_taumufhat_sigmamughat)**alpha

        
        return tau_, sigma_
    """
    if sm_gap < 0.5 * sm_gap_init:

        # Idea based on trying values and monitoring the convergence rate
        rho = max(0, sm_gap / sm_gap_init) ** (1/nb_iterations)  # using the new step size
        tau = primal_step_size[0]
        sigma = dual_step_size[0]
        if memory is None:
            up_down = 1
            memory = [rho, up_down]
        # history is a dictionary with values of tau as keys and the worst rate observed so far for these values
        if history == None:
            history = {tau:rho}  # initialization happens here
        else:
            if tau in history.keys():
                history[tau] = max([history[tau], rho])  # we store the worst case observed so far
            else:
                history[tau] = rho

        rho_ref = memory[0]  # using the older step size
        up_down = memory[1]

        if rho > rho_ref:
            # the new step size seems worse than the previous one: revert gear
            up_down = -up_down
        dual_step_size[:] = np.array(dual_step_size) * (alpha ** up_down)
        tmp_Lf = pb.gamma_step_size / np.array(primal_step_size) - pb.Lf
        tmp_Lf = tmp_Lf * (alpha ** up_down)
        primal_step_size[:] = pb.gamma_step_size / (pb.Lf + tmp_Lf)

        memory = [rho, up_down]

        print("step size changing", tau, "->", primal_step_size[0], sigma, "->", dual_step_size[0], ", 1-rho_ref = ", 1-rho_ref, "1-rho = ", 1-rho, "ratio = ", (1-rho) / max(1e-15, 1-rho_ref))

    return memory, history


def adaptive_steps_residual_norm(pb, x, Sy, prox_y, prox_y_cpy, rf,
                                 rhx, rQ, primal_step_size, dual_step_size, smoothed_gap, change_in_x, change_in_y,
                                 nb_iterations, memory, history=None, alpha=1.5, goldstein_warm_up=True):

    verbose = True
    # Idea based on trying values and monitoring the convergence rate in residual norm
    if memory is None:
        memory_goldstein = None
        if goldstein_warm_up == True:
            memory_goldstein, _ = adaptive_steps_goldstein(pb, x, Sy, prox_y,
                prox_y_cpy, rf, rhx, rQ, primal_step_size, dual_step_size,
                change_in_x, change_in_y,
                nb_iterations, memory_goldstein, history=[], alpha=alpha)
        rho_ref = 1
        up_down = 1
        counter = 1
        first = 0
        second = 0
        rates_history = [[1,1]]
        changes_in_iterates = [[change_in_x, change_in_y]]
        taus = [primal_step_size[0]]
        history = None
    else:
        memory_goldstein = memory[0]
        changes_in_iterates = memory[1]
        changes_in_iterates.append([change_in_x, change_in_y])
        rates_history = memory[2]  # [raw_value, filtered_value]
        first = memory[3][0]
        second = memory[3][1]
        rho_ref = memory[4]  # using the older step size
        up_down = memory[5]
        counter = memory[6]
        taus = memory[7]
        taus.append(primal_step_size[0])
        if counter > nb_iterations:
            counter = 0
            first = 0
            second = 0

        if memory_goldstein != None and memory_goldstein[-1] > 1e-2:
            # not much will change if alpha_goldstein < 0.01
            #   (maximal additional factor of change in stepsizes is 1.22 when eta = 0.95)
            tmp = memory_goldstein[-1]
            memory_goldstein, _ = adaptive_steps_goldstein(pb, x, Sy, prox_y, prox_y_cpy, rf,
                rhx, rQ, primal_step_size, dual_step_size, change_in_x, change_in_y,
                nb_iterations, memory_goldstein, history=[], alpha=2)
            if tmp != memory_goldstein[-1]:
                # the step sizes have changed
                counter = 0
                first = 0
                second = 0
                if verbose == 1:
                    print('goldstein step', primal_step_size[0])

        if counter <= 1:
            rates_history.append([np.nan, np.nan])  # Not enough data to measure
        if counter > 1:
            res_prev = changes_in_iterates[-2][0] ** 2 \
                + changes_in_iterates[-2][1] ** 2
            res = change_in_x ** 2 + change_in_y ** 2
            # the step sizes are included in this quantity when using pure_cd
            rho = np.sqrt(res / res_prev)

            if rho > 1:
                print("problem in residuals: they should decrease. ", counter, res, res_prev)
                counter = 0
                first = 0
                second = 0
            if counter == 2:
                rate_filtered = np.nan
            else:
                rate_filtered = max([rho, rates_history[-1][0], rates_history[-2][0]])
            rates_history.append([rho, rate_filtered])

        if counter > 4:
            res_far = changes_in_iterates[
                max(-len(changes_in_iterates), -counter-1)][0] ** 2 \
                + changes_in_iterates[
                    max(-len(changes_in_iterates), -counter-1)][1] ** 2
            if counter > 10:
                # we use filtered rate history to find cycles.
                if first > 0 and second == 0:
                    if np.sign(rates_history[-2][1] - rates_history[-3][1]) \
                       * np.sign(rates_history[-1][1] - rates_history[-2][1]) == -1:
                        second = counter
                        # TODO quadratic interpolation     
                if first == 0:
                    if np.sign(rates_history[-2][1] - rates_history[-3][1]) \
                       * np.sign(rates_history[-1][1] - rates_history[-2][1]) == -1:
                        first = counter
                if first > 0 and second > 0:
                    if second <= first + 3:
                        # spurious cycle, this is just noise
                        first = 0
                        second = 0
                    if res >= 0.6 * res_far:
                        # let's wait for a new cycle:
                        first = second
                        second = 0
                    else:
                        # we have detected a cycle and suppose there is a conjugate pair of principal eigenvalues
                        midpoint = (first + second) // 2
                        rho = rates_history[-counter+midpoint][1]

            rho_m1 = rates_history[-2][0]
            rho_m2 = rates_history[-3][0]

            if rho < 1 and rho_m1 < 1 and ((first > 0 and second > 0) or \
                (abs((1-rho)/(1-rho_m1)-1.) < 1e-3 and abs(rho - 2 * rho_m1 + rho_m2) / (1-rho_m1) < 1e-5)) \
               and res < 0.9 / alpha * res_far:
                # stabilization of the rate and sufficient decrease
                tau = primal_step_size[0]
                sigma = dual_step_size[0]
                if history == None:
                    history = {tau:rho}  # initialization happens here
                else:
                    if tau in history.keys():
                        history[tau] = max([history[tau], rho])  # we store the worst case observed so far
                    else:
                        history[tau] = rho

                if 1-rho < (1-rho_ref):
                    # the new step size seems worse than the previous one: revert gear
                    up_down = -up_down
                dual_step_size[:] = np.array(dual_step_size) * (alpha ** up_down)
                tmp_Lf = pb.gamma_step_size / np.array(primal_step_size) - pb.Lf
                tmp_Lf = tmp_Lf * (alpha ** up_down)
                primal_step_size[:] = pb.gamma_step_size / (pb.Lf + tmp_Lf)
                if verbose == 1:
                    print("step size changing", tau, "->", primal_step_size[0], sigma, "->", dual_step_size[0], ", 1-rho_ref = ", 1-rho_ref, "1-rho = ", 1-rho, "ratio = ", (1-rho) / max(1e-15, 1-rho_ref), ", cycle detection", first, second, ", convergence", abs(rho/rho_m1-1), abs(rho - 2 * rho_m1 + rho_m2) / rho_m1, ",sufficient decrease", res / (0.6 * res_far))

                rho_ref = rho
                counter = 0
                first = 0
                second = 0

            # print(counter, rho, rho_m1, rho_m2, rho_ref, first, second)
    counter = counter + 1
    memory = [memory_goldstein, changes_in_iterates, rates_history,
              [first, second], rho_ref, up_down, counter, taus]
    return memory, history


def estimate_rate_iid(instant_rates):
    log_rho = np.log(instant_rates)
    if len(instant_rates) > 1:
        mean = np.mean(log_rho)
        std = np.std(np.log(instant_rates)) / np.sqrt(len(instant_rates))
    else:
        mean = -1e-70
        std = 1000
    return mean, std


def estimate_rate_ar(instant_rates):
    if len(instant_rates) == 0:
        return -1e-70, 1000
    
    log_rho = np.log(instant_rates)
    ar_order = 1
    if len(instant_rates) > 2:
        A = np.hstack([np.ones((len(instant_rates)-1,1)), log_rho[0:-1].reshape(-1,1)])
        b = log_rho[1:]
        a = np.linalg.lstsq(A, b, rcond=None)[0]
        sigma = np.linalg.norm(A @ a - b) / np.sqrt(len(b))
        var = sigma**2 / (1-a[1])**2            
        if np.abs(a[1]) >= 1:
            var = 1e6
        return a[0] / (1 - a[1]), np.sqrt(var) / np.sqrt(len(b))
    else:
        return np.mean(log_rho), 1000


def adaptive_steps_stochastic(pb, x, Sy, prox_y, prox_y_cpy, rf,
                              rhx, rQ, primal_step_size, dual_step_size, smoothed_gap, change_in_x, change_in_y,
                              nb_iterations, memory, history=None, delta = 0.4, goldstein_warm_up=True,
                              estimate_rate=estimate_rate_ar):

    alpha = 2  # having a power of 2 helps for the keys of the history list
    # we need alpha delta < 1
    delta_warmup = 0.01
    verbose = 1

    log_rho = np.nan
    std = np.nan
    smgap = max([1e-50, smoothed_gap])  # not really what I wanted but this quantity is already computed...

    if memory is None:
        memory = []
        memory_goldstein = None
        if goldstein_warm_up == True:
            memory_goldstein, _ = adaptive_steps_goldstein(pb, x, Sy, prox_y,
                                                           prox_y_cpy, rf, rhx, rQ, primal_step_size, dual_step_size,
                                                           smoothed_gap, change_in_x, change_in_y,
                                                           nb_iterations, memory_goldstein, history=[], alpha=alpha,
                                                           eta=0.9)
        
        tau0 = primal_step_size[0]
        sigma0 = dual_step_size[0]
        history = {dual_step_size[0]: []}
        up_down = 1
        sigma_checked = sigma0
        if goldstein_warm_up == True:
            sigma_balance = sigma0
            do_residual_balance = True
            sigma_tentative = None
        else:
            sigma_balance = None
            do_residual_balance = False
            sigma_tentative = sigma_checked * (alpha ** up_down)
        counter = 0
    else:
        counter = memory[-1][4]
        memory_goldstein = memory[-1][5]
        sigma_checked, sigma_balance, sigma_tentative = memory[-1][6]
        do_residual_balance = memory[-1][7]

        if do_residual_balance == True:
            tmp = memory_goldstein[-1]
            if tmp > 0.01:
                memory_goldstein, _ = adaptive_steps_goldstein(pb, x, Sy, prox_y, prox_y_cpy, rf,
                                                           rhx, rQ, primal_step_size, dual_step_size, smoothed_gap,
                                                           change_in_x, change_in_y,
                                                               nb_iterations, memory_goldstein, history=[], alpha=2, eta=0.95)
            if tmp != memory_goldstein[-1]:
                # the step sizes have changed
                if verbose == 1:
                    print('goldstein step', dual_step_size[0], ' alpha =', memory_goldstein[-1])

        history = memory[-1][0]
        up_down = memory[-1][2]
        sm_gap_m1 = memory[-1][3]
        if sm_gap_m1 == smgap:  # We do not really control when the smoothed gap is updated
            memory[-1][5] = memory_goldstein
            return memory, history  # no data
        else:
            instant_rate = smgap / sm_gap_m1
        sigma0 = dual_step_size[0]
        tau0 = primal_step_size[0]
        if sigma0 in history.keys():
            history[sigma0].append(instant_rate)
            if len(history[sigma0]) > len(memory) // 4:
                history[sigma0].pop(0)
        else:
            history[sigma0] = [instant_rate]
            
        instant_rates = np.array(history[sigma0])
            
        log_rho, std = estimate_rate(instant_rates)
        rho = np.exp(log_rho)
        
        if counter > 3:
            sm_gap_previous = memory[-counter][3]
            if do_residual_balance == True:
                if smgap < delta_warmup * sm_gap_previous \
                       or memory_goldstein[-1] <= 0.01:
                    do_residual_balance = False
                    sigma_balance = sigma0
                    if sigma_balance > sigma_checked:
                        up_down = 1
                        if sigma_balance < sigma_checked * alpha ** up_down:
                            sigma_tentative = sigma_checked * alpha ** up_down
                        else:
                            sigma_tentative = sigma_balance
                    if sigma_balance <= sigma_checked:
                        up_down = -1
                        if sigma_balance > sigma_checked * alpha ** up_down:
                            sigma_tentative = sigma_checked * alpha ** up_down
                        else:
                            sigma_tentative = sigma_balance
                        
                    print('goldstein warm up finished', sigma_checked, sigma_balance, sigma_tentative)

            else:
                instant_rates_checked = np.array(history[sigma_checked])
                log_rho_checked, std_checked = estimate_rate(instant_rates_checked)

                if sigma_tentative in history.keys():
                    instant_rates_tentative = np.array(history[sigma_tentative])
                    log_rho_tentative, std_tentative = estimate_rate(instant_rates_tentative)
                else:
                    log_rho_tentative = log_rho_checked
                    std_tentative = 1000

                prob = 0.5 * sps.erfc((-log_rho_checked + log_rho_tentative)
                                             / np.sqrt(std_checked**2 + std_tentative**2) / np.sqrt(2))

                if smgap < delta * sm_gap_previous or counter > 100:
                    if counter < len(memory):
                        if prob <= 0.55 and prob >= 0.45:
                            # not enough information: try to increase knowledge
                            if sigma0 == sigma_tentative:
                                sigma_next = sigma_checked
                            else:  # sigma0 = sigma_checked
                                sigma_next = sigma_tentative
                        elif prob < 0.45:
                            # the tentative is a failure
                            up_down = -up_down
                            sigma_tentative = sigma_checked * (alpha ** up_down)
                            if sigma0 == sigma_checked:
                                sigma_next = sigma_tentative
                            else:
                                sigma_next = sigma_checked
                                
                        else:  # prob > 0.55:
                            # the tentative is a success
                            sigma_checked = sigma_tentative
                            sigma_tentative = sigma_checked * (alpha ** up_down)
                            sigma_next = sigma_tentative

                        sigma_previous = sigma0
                        change = sigma_next / sigma0
                        dual_step_size[:] = np.array(dual_step_size) * change
                        tmp_Lf = pb.gamma_step_size / np.array(primal_step_size) - pb.Lf
                        tmp_Lf = tmp_Lf * change
                        primal_step_size[:] = pb.gamma_step_size / (pb.Lf + tmp_Lf)
                        print('changing step sizes', sigma0, '->', dual_step_size[0], '(prob = ', prob, ')', '1-rho_c = ', 1-np.exp(log_rho_checked), '1-rho_t = ', 1-np.exp(log_rho_tentative), std_checked, std_tentative, sigma_checked, sigma_tentative)
                        counter = 0
        counter += 1

    memory.append([history, [primal_step_size[0], dual_step_size[0]], up_down, smgap, counter, memory_goldstein, [sigma_checked, sigma_balance, sigma_tentative], do_residual_balance])

    return memory, history


def dual_run_adaptive_steps(
        pb, x, Sy, prox_y, prox_y_cpy, rf,
        rhx, rQ, primal_step_size, dual_step_size, sm_gap_init, sm_gap, change_in_x, change_in_y, nb_iterations, memory, history=None, alpha=2):
    
    if sm_gap < 0.5 * sm_gap_init:
        if memory is None:
            # We initialize the memory
            up_down = 1
            type_of_iteration = 0
            memory = copy.deepcopy([np.array(x), np.array(Sy), np.array(prox_y), np.array(prox_y_cpy), np.array(rf),
                                    np.array(rhx), np.array(rQ), np.array(primal_step_size), np.array(dual_step_size), sm_gap,
                                    nb_iterations, alpha, up_down, type_of_iteration])
            cancel_restart = False
        else:
            type_of_iteration = memory[-1]
            if type_of_iteration == 0:
                # We need to run again with different step sizes
                x0, Sy0, prox_y0, prox_y_cpy0, rf0, \
                    rhx0, rQ0, primal_step_size0, dual_step_size0, \
                    sm_gap0, nb_iterations0, alpha, up_down, type_of_iteration = copy.deepcopy(memory)
                type_of_iteration = 1

                memory = copy.deepcopy([np.array(x), np.array(Sy), np.array(prox_y), np.array(prox_y_cpy), np.array(rf),
                                        np.array(rhx), np.array(rQ), np.array(primal_step_size), np.array(dual_step_size),
                                        sm_gap, nb_iterations, 
                                        alpha, up_down, type_of_iteration])

                # Change current point but not the pointers to try again with new step sizes
                x[:] = x0.copy()
                Sy[:] = Sy0.copy()
                prox_y[:] = prox_y0.copy()
                prox_y_cpy[:] = prox_y_cpy0.copy()
                rf[:] = rf0.copy()
                rhx[:] = rhx0.copy()
                rQ[:] = rQ0.copy()

                dual_step_size[:] = np.array(dual_step_size) * (alpha ** up_down)
                tmp_Lf = pb.gamma_step_size / np.array(primal_step_size) - pb.Lf
                tmp_Lf = tmp_Lf * (alpha ** up_down)
                primal_step_size[:] = pb.gamma_step_size / (pb.Lf + tmp_Lf)

                cancel_restart = True

            else:  # type_of_iteration == 1:

                # We need to compare both runs
                sm_gap0 = memory[-5]
                nb_iterations0 = memory[-4]
                rho = max(0, sm_gap / sm_gap_init) ** (1/nb_iterations)  # using the new step size
                rho0 = max(0, sm_gap0 / sm_gap_init) ** (1/nb_iterations0)  # using the old step size
                print("smgap", sm_gap, "nb_it", nb_iterations, "smgap0", sm_gap0, "nb_it0", nb_iterations0)

                if rho0 > rho:
                    # the new step size is better, we throw away the old memory (except up_down) and continue
                    up_down = memory[-2]
                    memory = copy.deepcopy([np.array(x), np.array(Sy), np.array(prox_y), np.array(prox_y_cpy), np.array(rf),
                                        np.array(rhx), np.array(rQ), np.array(primal_step_size), np.array(dual_step_size),
                                            sm_gap, nb_iterations, alpha, up_down, 0])
                else:
                    # the old step size was better: revert gear and forget last run
                    # Question: do we need to reset the current point also?
                    up_down = - memory[-2]
                    memory[-2] = up_down
                    memory[-1] = 0

                # history is a dictionary with values of tau as keys and the worst rate observed so far for these values
                tau = primal_step_size[0]
                if history == None:
                    history = [[tau, rho]]  # initialization happens here
                else:
                    history.append([tau, rho])
                cancel_restart = False

                print("step size changing: tau = ", primal_step_size[0], " sigma = ", dual_step_size[0], ", up_down = ", memory[-2], "1-rho = ", 1-rho, "1-rho0 = ", 1-rho0, "ratio = ", (1-rho)/(1-rho0))

    return memory, history, cancel_restart


def adaptive_steps_goldstein(pb, x, Sy, prox_y, prox_y_cpy, rf,
                             rhx, rQ, primal_step_size, dual_step_size, smoothed_gap, change_in_x, change_in_y,
                             nb_iterations, memory, history=None, alpha=2, eta=0.98):
    # Tom Goldstein, Min Li, Xiaoming Yuan, Ernie Esser, and Richard Baraniuk. Adaptive primal-dual hybrid
    #   gradient methods for saddle-point problems. arXiv preprint: arXiv 1305.0546, 2015.
    # For usage in the original way, this should be used with fixed_restart_period='just_check' and restart_period=1
    # we wrote it for Latafat's version of PDHG, so the definition of the primal and dual residuals are adapted slightly.

    alpha0 = 1. / (alpha * eta)
    Delta = 1.5
    if memory is None:
        history = [primal_step_size[0]]
        alpha_ = alpha0
        up_down = 2  # code for nothing happened yet
        # we will update alpha before updating the step size in this code,
        # so we correct this by dividing by eta here.
    else:
        #x0, Sy0, prox_y0, prox_y_cpy0, \
        #    nb_iterations0, up_down0, alpha_ = copy.deepcopy(memory)
        alpha_ = memory[0]

        primal_residual = change_in_x  # np.linalg.norm((x0 - np.array(x))/primal_step_size[0], 1)
        dual_residual = change_in_y  # np.linalg.norm((Sy0 - np.array(Sy))/dual_step_size[0], 1)

        up_down = 0
        if primal_residual > Delta * dual_residual:
            up_down = 1
            alpha_ = alpha_ * eta
        elif dual_residual > Delta * primal_residual:
            up_down = -1
            alpha_ = alpha_ * eta
        # else: keep up_down = 0
        
        dual_step_size[:] = np.array(dual_step_size) * ((1-alpha_) ** up_down)
        tmp_Lf = pb.gamma_step_size / np.array(primal_step_size) - pb.Lf
        tmp_Lf = tmp_Lf * ((1-alpha_) ** up_down)
        primal_step_size[:] = pb.gamma_step_size / (pb.Lf + tmp_Lf)

        history.append(primal_step_size[0])

    memory = [alpha_]  #copy.deepcopy([np.array(x), np.array(Sy), np.array(prox_y),
                       #     np.array(prox_y_cpy), nb_iterations,
                       #     up_down, alpha_])

    return memory, history
